package descartes.kieker4dql.dqlconnector;

import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.layout.GridDataFactory;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.FontData;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Text;
import org.eclipse.ui.PlatformUI;

import edu.kit.ipd.descartes.solving.ValidationError;
import edu.kit.ipd.descartes.solving.ValidationErrorSource;

public class ValidationErrorDialog extends MessageDialog {
	Collection<ValidationError> errors;
	
	public ValidationErrorDialog(Shell parentShell, Collection<ValidationError> errors) {
		super(parentShell, "ValidationError", null, "",
				MessageDialog.ERROR, new String[] {  "OK" }, 0);
		  this.errors = errors;
	}

	@Override
	public Control createDialogArea(Composite parent) {
		addImage(parent);
		Composite messageArea = createMessageComposite(parent);
		createHeadline(messageArea, "Validation failed!");
		createFullRowText(messageArea, "The DML-Files you provided contained invalid data."
				+ " Please fix the following issues and try again.");
		
		Composite nonScroll = createNonScrollContent(messageArea);
		if (getHeight(nonScroll) > 400) {
			nonScroll.dispose();
			createScrollContent(messageArea);
		}
		return parent;
	}
	
	private void createScrollContent(Composite composite) {
		final ScrolledComposite sc = new ScrolledComposite(composite, SWT.V_SCROLL);
		sc.setExpandHorizontal(true);
		sc.setExpandVertical(true);
		Composite content = new Composite(sc, SWT.NONE);
		sc.setContent(content);
		GridData data = new GridData(SWT.FILL, SWT.FILL, true, true);
		data.heightHint = 400;
		sc.setLayoutData(data);
		content.setLayout(new GridLayout(2, false));
		createContent(content);
		sc.setMinSize(content.computeSize(SWT.DEFAULT, SWT.DEFAULT));
		
	}

	private int getHeight(Composite composite) {
		return composite.computeSize(SWT.DEFAULT, SWT.DEFAULT).y;
	}
	
	private Composite createNonScrollContent(Composite parent) {
		Composite composite = new Composite(parent, SWT.NONE);
		composite.setLayout(new GridLayout(2, false));
		createContent(composite);
		return composite;
	}
	
	private void createContent(Composite parent) {
		Map<ValidationErrorSource, List<String>> errorMap = getErrorMap();
		for (Map.Entry<ValidationErrorSource,List<String>> entry: errorMap.entrySet()) {
			createSubHeadline(parent, entry.getKey().getName());
			for (String message: entry.getValue())
				createBulletPoint(parent, message);
		}
	}
	
	private void createBulletPoint(Composite parent, String text) {
		Text bullet = new Text(parent, SWT.READ_ONLY);
		bullet.setText("-");
		GridData gridData = new GridData(GridData.VERTICAL_ALIGN_BEGINNING, GridData.HORIZONTAL_ALIGN_BEGINNING, false, false);
		bullet.setLayoutData(gridData);
		Text bulletText = new Text(parent, SWT.WRAP | SWT.READ_ONLY);
		GridData gd = new GridData ();
		gd.widthHint = 700;
		bulletText.setLayoutData (gd);
		bulletText.setText(insertNewLines(text, bulletText.getFont(), 700));
	}

	private Text createSubHeadline(Composite parent, String text) {
		Text subHeadLine = createFullRowText(parent, text);
		Font font = new Font(PlatformUI.getWorkbench().getDisplay(),
				new FontData("Headline Font", 12, SWT.BOLD));
		subHeadLine.setFont(font);
		return subHeadLine;
	}
	
	private Text createHeadline(Composite parent, String text) {
		Text headLine = createFullRowText(parent, text);
		Font font = new Font(PlatformUI.getWorkbench().getDisplay(),
				new FontData("Headline Font", 16, SWT.BOLD));
		headLine.setFont(font);
		return headLine;
	}
	
	private Text createFullRowText(Composite parent, String text) {
		Text label = new Text(parent, SWT.READ_ONLY);
		label.setText(text);
		GridData gridData = new GridData(GridData.FILL, GridData.FILL, true, true);
		gridData.horizontalSpan = 2;
		label.setLayoutData(gridData);
		return label;
	}
	
	private Composite createMessageComposite(Composite parent) {
		Composite panel = new Composite(parent, SWT.NONE);
		GridData gd = new GridData(GridData.FILL_BOTH);
		gd.minimumWidth = 700;
		panel.setLayoutData(gd);
		GridLayout layout = new GridLayout();
		layout.marginHeight = 0;
		layout.marginWidth = 0;
		layout.numColumns = 2;
		panel.setLayout(layout);
		return panel;
	}

	private void addImage(Composite composite) {
		Image image = getImage();
		if (image != null) {
			imageLabel = new Label(composite, SWT.NULL);
			image.setBackground(imageLabel.getBackground());
			imageLabel.setImage(image);
			GridDataFactory.fillDefaults().align(SWT.CENTER, SWT.BEGINNING)
					.applyTo(imageLabel);
		}
	}
	
	private Map<ValidationErrorSource, List<String>> getErrorMap() {
		Map<ValidationErrorSource, List<String>> map = new HashMap<ValidationErrorSource, List<String>>();
		for (ValidationError error: errors) {
			if (!map.containsKey(error.getSource()))
				map.put(error.getSource(), new LinkedList<String>());
			map.get(error.getSource()).add(error.getMessage());
		}
		return map;
	}
	
	private String insertNewLines(String text, Font f, int width) {
		String edited = "";
		while (text.length() != 0) {
			int initialCutoff = getInitialCutoff(text, f, width);
			int newCutoff = adjustCutoffForSmallWords(text, initialCutoff);
			edited += text.substring(0, newCutoff);
		    if (newCutoff != text.length())
		    	edited += "\n";
			text = text.substring(newCutoff);
		}
		return edited;
	}
	
	private int adjustCutoffForSmallWords(String text, int initialCutoff) {
		int start = getPosOfFirstLetter(text, initialCutoff);
		int end = getPosOfNextWhitespace(text, initialCutoff);
		int wordLength = end - start;
		if (end != -1 && wordLength < 20)
			return start;
		return initialCutoff;
	}

	private int getPosOfNextWhitespace(String text, int initialCutoff) {
		int pos = text.indexOf(" ", initialCutoff);
		return pos;
	}

	private int getPosOfFirstLetter(String text, int initialCutoff) {
		int pos = text.lastIndexOf(" ", initialCutoff) + 1;
		return pos;
	}

	private int getInitialCutoff(String text, Font f, int width) {
		String buffer = "";
		int pos = 0;
		while (getWidth(buffer, f) < width && pos < text.length()) {
			buffer += text.charAt(pos);
			pos++;
		}
		if (getWidth(buffer, f) >= width) {
			pos--;
		}
		return pos;
	}
	
	private int getWidth(String s, Font f) {
		Text l = new Text(getShell(), SWT.NONE);
		l.setFont(f);
		GC gc = new GC(l);
		Point size = gc.textExtent(s);
		gc.dispose();
		l.dispose();
		return size.x;
	}
}
